/*********************************************************************NVMH3****
File:  $Id: //devrel/Playpen/kbjorke/HLSL/PointShadowAimRad8_twin.fx#2 $

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

<scene_commands>
<!-- make shadow map "ShadMap" -->
<setcolortarget texture="ShadMap"/>
<setdepthstenciltarget texture="ShadDepthTarget"/>
	<clear color="1.0,1.0,1.0,0" depth="1.0"/>
	<settechnique name="MakeShadow"/>
	<draw type="objects_override"/>
<setcolortarget/>
<setdepthstenciltarget/>
	<clear color="0.0,0.0,0.0,0" depth="1.0"/>
	<settechnique name="UseShadow"/>
	<draw type="objects_override"/>
	<settechnique name="ShareShadow"/>
	<draw type="objects_override"/>
	<!-- <settechnique name="ShowShadow"/>
	<draw type="buffer"/> -->
</scene_commands>

******************************************************************************/

#ifndef SHADOW_SIZE
#define SHADOW_SIZE 512
#endif /* SHADOW_SIZE */

#ifndef SHADOW_FMT
//#define SHADOW_FMT "A8R8G8B8"
#define SHADOW_FMT  "a16b16g16r16f"
#endif /* SHADOW_FMT */

#include "include/nvMatrix.fxh"
#include "include/Quad.fxh"

///////////////////////////////////////////////////////////////
/// TWEAKABLES ////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

float3 UpVector <
	string UIName = "Up Vector";
	// bool Normalized = 1;
> = { 0.0f, 0.0f, 1.0f};

////////////////////////////////////////////// spot light

float3 PointLightPos : POSITION <
	string UIName = "Light Pos";
	string Object = "PointLight";
	string Space = "World";
> = {-1.0f, 1.0f, 0.0f};

float3 PointLightColor : Specular <
	string UIName = "Light Color";
	string Object = "PointLight";
	string UIWidget = "Color";
> = {0.8f, 1.0f, 0.4f};

float PointLightIntensity <
	string UIName = "Light Intensity";
	string UIWidget = "slider";
	float UIMin = 0.0;
	float UIMax = 200;
	float UIStep = 0.1;
> = 1;

float3 Pos2 : POSITION <
	string UIName = "Light Pos 2";
	string Object = "PointLight";
	string Space = "World";
> = {-1.0f, 1.0f, 0.0f};

float3 LightColor2 : Specular <
	string UIName = "Light Color 2";
	string Object = "PointLight";
	string UIWidget = "Color";
> = {0.8f, 1.0f, 0.4f};

float Intensity2 <
	string UIName = "Light Intensity 2";
	string UIWidget = "slider";
	float UIMin = 0.0;
	float UIMax = 200;
	float UIStep = 0.1;
> = 1;

/////////////// Shadow Parameters

float3 ShadowCenter : POSITION <
	string UIName = "Shadow is centered here";
	string Object = "PointLight";	// hack for now -- no nulls
	string Space = "World";
> = {0.0f, 0.0f, 0.0f};

float ShadRadius <
    string UIWidget = "slider";
    float UIMin = 0.01;
    float UIMax = 15.0;
    float UIStep = 0.01;
    string UIName = "Shadow Radius";
> = 5.0;

float ShadNear <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 4.0;
    float UIStep = 0.01;
    string UIName = "Shadow Near Fudge";
> = 0.0;

float ShadFar <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 4.0;
    float UIStep = 0.01;
    string UIName = "Shadow Far Fudge";
> = 0.0;

float ShadBias <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 0.1;
    float UIStep = 0.001;
    string UIName = "Shadow Bias";
> = 0.01;

////////////////////////////////////////////// ambient light

float3 AmbiLightColor : Ambient
<
    string UIName = "Ambient Light Color";
> = {0.07f, 0.07f, 0.07f};

////////////////////////////////////////////// surface

float3 SurfColor : Diffuse
<
    string UIName = "Surface Color";
	string UIWidget = "Color";
> = {1.0f, 0.7f, 0.3f};

float Kd
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.5;
    float UIStep = 0.01;
    string UIName = "Diffuse";
> = 1.0;

float Ks
<
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.5;
    float UIStep = 0.01;
    string UIName = "Specular";
> = 1.0;


float SpecExpon : SpecularPower
<
    string UIWidget = "slider";
    float UIMin = 1.0;
    float UIMax = 128.0;
    float UIStep = 1.0;
    string UIName = "Specular power";
> = 12.0;

////////////////////////////////////////////////////////
/// TEXTURES ///////////////////////////////////////////
////////////////////////////////////////////////////////

// fixed-size shadow texture

texture ShadMap : RENDERCOLORTARGET < 
    // float2 ViewPortDimensions = {1.0,1.0};
	int Width = SHADOW_SIZE;
	int Height = SHADOW_SIZE;
    // int MIPLEVELS = 1;
    string format = (SHADOW_FMT);
    string UIWidget = "None";
>;

sampler ShadSampler = sampler_state {
    texture = <ShadMap>;
    AddressU  = BORDER;        
    AddressV  = BORDER;
    MIPFILTER = NONE;
    MINFILTER = LINEAR;
    MAGFILTER = LINEAR;
};

texture ShadDepthTarget : RENDERDEPTHSTENCILTARGET < 
	int Width = SHADOW_SIZE;
	int Height = SHADOW_SIZE;
    string format = "D24S8";
    string UIWidget = "None";
>;

/////////////////////////////////////////////////////////////

#ifdef USE_NORMALIZATION_CUBEMAP
#include "include/normalize.fxh"
// only for pixel shaders....

/*
half3 my_normalize(half3 v)
{
	half3 v2 = texCUBE(NormalizeSampler,v);
	return (2*(v2-0.5));
}
half4 my_normalize(half4 v)
{
	half3 v2 = texCUBE(NormalizeSampler,v.xyz);
	return half4((2*(v2-0.5)),1);
}
*/

#define NORM my_normalize
#else /* !USE_NORMALIZATION_CUBEMAP */
#define NORM normalize
#endif /* !USE_NORMALIZATION_CUBEMAP */

/************* "UN-TWEAKABLES," TRACKED BY CPU APPLICATION **************/

float4x4 WorldITXf : WorldInverseTranspose <string UIWidget="None";>;
float4x4 WorldViewProjXf : WorldViewProjection <string UIWidget="None";>;
float4x4 WorldXf : World <string UIWidget="None";>;
float4x4 ViewIXf : ViewInverse <string UIWidget="None";>;
float4x4 WorldViewITXf : WorldViewInverseTranspose <string UIWidget="None";>;
float4x4 WorldViewXf : WorldView <string UIWidget="None";>;
float4x4 ViewXf : View <string UIWidget="None";>;
float4x4 ViewITXf : ViewInverseTranspose <string UIWidget="None";>;

////////////////////////////////////////////////////////////////////////////
/// SHADER CODE BEGINS /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

/*********************************************************/
/************* SHARED DATA STRUCT ************************/
/*********************************************************/

/* data from application vertex buffer */
struct appdata {
    float3 Position	: POSITION;
    float4 UV		: TEXCOORD0;
    float4 Normal	: NORMAL;
};

///////////////////////////////////////////////////////////////////////////////
/// Basic Version Does Almost All Calcs in the Pixel Shader ///////////////////
///////////////////////////////////////////////////////////////////////////////

struct vertexOutput {
    float4 HPosition	: POSITION;
    float2 UV		: TEXCOORD0;
    float3 LightVec	: TEXCOORD1;
    float3 WNormal	: TEXCOORD2;
    float3 WView		: TEXCOORD3;
    float4 LP		: TEXCOORD4;
};

/*********************************************************/
/*********** VM funcs ************************************/
/*********************************************************/

float4x4 shadowProjXf()
{
	float3 aimVec = (ShadowCenter-PointLightPos);
	float d = length(aimVec);
	float shadCone = asin(ShadRadius/length(aimVec));
	float3 aimN = normalize(aimVec);
	float4x4 spotXf = nv_spot_xf(PointLightPos,aimN,UpVector);
	float shadHither = max(0.01,(d-(ShadRadius+ShadNear)));
	float shadYon = d+ShadRadius+ShadFar;
	float4x4 projXf = nv_spot_proj_xf(spotXf,
									shadCone,
									shadHither,
									shadYon);
	return projXf;
}

////////////////////////////////////////////////////////////////////////////////
/// Vertex Shaders /////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// from lamp POV
vertexOutput shadVS(appdata IN,
					uniform float4x4 ShadowXf)
{
    vertexOutput OUT = (vertexOutput)0;
    OUT.WNormal = mul(IN.Normal,WorldITXf).xyz;
    float4 Po = float4(IN.Position.xyz,(float)1.0);	// object coordinates
    float4 Pw = mul(Po,WorldXf);			// world coordinates
    float4 Pl = mul(Pw,ShadowXf);  // "P" in light coords

    OUT.LP = Pl;		// view coords

    OUT.WView = normalize(ViewIXf[3].xyz - Pw.xyz);	// obj coords
    OUT.HPosition = Pl; // mul(Po,WorldViewProjXf);	// screen clipspace coords
    OUT.UV = IN.UV.xy;
    OUT.LightVec = PointLightPos - Pw.xyz; // in world coords
    return OUT;
}

// from scene camera POV
vertexOutput mainCamVS(appdata IN,
					uniform float4x4 ShadowXf,
					uniform float3 LPos)
{
    vertexOutput OUT = (vertexOutput)0;
    OUT.WNormal = mul(IN.Normal,WorldITXf).xyz;
    float4 Po = float4(IN.Position.xyz,(float)1.0);	// object coordinates
    float4 Pw = mul(Po,WorldXf);			// world coordinates
    float4 Pl = mul(Pw,ShadowXf);  // "P" in light coords

    OUT.LP = Pl;		// view coords

    OUT.WView = normalize(ViewIXf[3].xyz - Pw.xyz);	// obj coords
    OUT.HPosition = mul(Po,WorldViewProjXf);	// screen clipspace coords
    OUT.UV = IN.UV.xy;
    OUT.LightVec = LPos - Pw.xyz; // in world coords
    return OUT;
}

/*********************************************************/
/*********** pixel shader ********************************/
/*********************************************************/

float4 makeShadPS(vertexOutput IN) : COLOR
{
    return float4(IN.LP.zzz/IN.LP.w,1);
}

float4 useShadowPS(vertexOutput IN,
	uniform float3 LColor,
	uniform float LIntens) : COLOR
{
    float3 Nn = NORM(IN.WNormal);
    float3 Vn = NORM(IN.WView);
    float fall = LIntens/dot(IN.LightVec,IN.LightVec);
    float3 Ln = normalize(IN.LightVec);
    float3 Hn = NORM(Vn + Ln);
    float hdn = dot(Hn,Nn);
    float ldn = dot(Ln,Nn);
    float4 litVec = lit(ldn,hdn,SpecExpon);
    ldn = litVec.y * fall;
    float3 ambiContrib = SurfColor * AmbiLightColor;
    float3 diffContrib = SurfColor*(Kd*ldn * LColor);
    float3 specContrib = ((ldn * litVec.z * Ks) * LColor);
    float3 result = diffContrib + specContrib;
    // return float4(result,1);
    float2 nuv = float2(.5,-.5)*IN.LP.xy/IN.LP.w + float2(.5,.5);
    float shad = tex2D(ShadSampler,nuv).x;
    float depth = (IN.LP.z/IN.LP.w) - ShadBias;
    //shad = ShadBias+ShadNear + shad*(ShadFar-ShadNear);
    float inShad = 1-(shad<depth);
    return float4((inShad*result)+ambiContrib,1);
    // return float4(inShad.xxx,1);
}

////////////////////////////////////////////////////////////////////
/// TECHNIQUES /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////

technique MakeShadow {
	pass shadPass {
		VertexShader = compile vs_2_0 shadVS(shadowProjXf());
		ZEnable = true;
		ZWriteEnable = true;
		CullMode = None;
		PixelShader = compile ps_2_a makeShadPS();
	}
}

technique UseShadow {
	pass drawPass {
		VertexShader = compile vs_2_0 mainCamVS(shadowProjXf(),PointLightPos);
		ZEnable = true;
		ZWriteEnable = true;
		CullMode = None;
		PixelShader = compile ps_2_a useShadowPS(PointLightColor,PointLightIntensity);
	}
}

technique ShareShadow {
	pass drawPass {
		VertexShader = compile vs_2_0 mainCamVS(shadowProjXf(),Pos2);
		ZEnable = true;
		ZWriteEnable = false;
		ZFunc = LessEqual;
		CullMode = None;
		AlphaBlendEnable = true;
		SrcBlend = one;
		DestBlend = one;
		PixelShader = compile ps_2_a useShadowPS(LightColor2,Intensity2);
	}
}
/***************************** eof ***/
